package uk.ac.ox.eolo.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.ConnectivityInspector;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.HermiT.model.Atom;
import org.semanticweb.HermiT.model.Individual;
import org.semanticweb.HermiT.model.Role;
import org.semanticweb.HermiT.model.Term;

import uk.ac.ox.cs.JRDFox.JRDFStoreException;
import uk.ac.ox.cs.JRDFox.store.TupleIterator;
import uk.ac.ox.eolo.URIUtils;
import uk.ac.ox.eolo.reasoner.DatalogReachability;
import uk.ac.ox.eolo.reasoner.EOLO;
import uk.ac.ox.eolo.reasoner.SoundChecker;

public class Skeleton {

	DirectedGraph<Term, DefaultEdge> skeleton;
	EOLO reasoner;
	ConjunctiveQuery sigma_q;	
	Map<Term, String> tau;
	
	
	
	public static Skeleton create(DirectedGraph<Term, DefaultEdge> g, EOLO r, ConjunctiveQuery q, Map<Term, String> tau) {
		Skeleton s = new Skeleton();
		s.skeleton = new DefaultDirectedGraph<Term, DefaultEdge>(DefaultEdge.class);
		for (Term v : g.vertexSet())
			s.skeleton.addVertex(v);
		for (Term v : g.vertexSet())
			for (Term t: g.vertexSet())
				if (g.containsEdge(v, t))
					s.skeleton.addEdge(v, t);
		s.reasoner = r;
		s.sigma_q =q;
		s.tau = tau;
		return s;
	}


	public void addEdge(Term term, Term key) {
		skeleton.addEdge(term, key);
	}
	
	
	public Set<Term> vertexSet() {
		return skeleton.vertexSet();
	}
	
	
	public boolean containsEdge(Term s, Term t) {
		return skeleton.containsEdge(s, t);
	}
	
	
	
	public boolean isForest() {
		return !(new CycleDetector<Term, DefaultEdge>(skeleton)).detectCycles();
	}

	/**
	 * Returns the root of t in the skeleton
	 * @param skeleton
	 * @param t
	 * @return
	 */
	public Term getRoot(Term t) {
		if (skeleton.incomingEdgesOf(t).isEmpty())
			return t;
		for (Term root : skeleton.vertexSet()) {
			if ((root instanceof Individual) || (root instanceof RootVariable))
				if ((new ConnectivityInspector<Term, DefaultEdge>(skeleton)).pathExists(root, t))
					return root;
		}
		return null;
	}
	
	
	
	public boolean reaches(Term s, Term t ) {
		if (skeleton.containsEdge(s, t))
			return true;
		for (Term u : skeleton.vertexSet()) {
			if (skeleton.containsEdge(u, t))
				return reaches(s,u);
		}
		return false;
	}
	
	
	public List<DefaultEdge> getPath(Term s, Term t) {
		if (skeleton.incomingEdgesOf(t).isEmpty())
			return null;
		List<DefaultEdge> path = new ArrayList<DefaultEdge>();
		if (skeleton.containsEdge(s, t)) {
			path.add(skeleton.getEdge(s, t));
			return path;
		}
		Term current = t;
		int i =0;
		while (!current.equals(s) && i < 10) {
			for (Term pred : skeleton.vertexSet()) {
				if (skeleton.containsEdge(pred, current)) {
					path.add(skeleton.getEdge(pred, current));
					current = pred;
				}
			}
			i++;
		}
		return path;
	}


	public Set<DefaultEdge> outgoingEdgesOf(Term s) {
		return skeleton.outgoingEdgesOf(s);
	}


	public DefaultEdge getEdge(Term s, Term t) {
		return skeleton.getEdge(s, t);
	}


	public Set<DefaultEdge> edgeSet() {
		return skeleton.edgeSet();
	}
	
	public  Labelling createLabelling() throws Exception {
		Labelling label = new Labelling();
		Map<DefaultEdge, Set<String>> L = new HashMap<DefaultEdge, Set<String>>();
		for (Atom a : sigma_q.getAtoms())
			if (a.getArity() > 1 && reasoner.isAuxSimple(a, tau, sigma_q.getPrefixes())) {
				DefaultEdge edge = this.getEdge(a.getArgument(0), a.getArgument(1));
				Set<String> roles = (L.containsKey(edge)) ? L.get(edge) : new HashSet<String>();
				roles.add(URIUtils.getRole((Role) a.getDLPredicate(), sigma_q.getPrefixes()));
				L.put(edge, roles);
			}
		label.L =L;
		return label;	
	}
	
	
	public class Labelling {
		Map<DefaultEdge, Set<String>> L;
		
		
		/**
		 * Adds role P to L(v', v) for each edge (v',v) in the path.
		 * @param L
		 * @param p
		 * @param path
		 * @throws Exception 
		 */
		public void put(String p, List<DefaultEdge> path) throws Exception {
			if (path == null)
				throw new Exception("path is empty");
			for (DefaultEdge e : path) {
				Set<String> roles = (L.containsKey(e)) ? L.get(e)	: new HashSet<String>();
				roles.add(p);
				L.put(e, roles);
			}
		}
		
		/**
		 * Check whether each edge in the skeleton conforms to L and the constraints. For each root variable x_root,
		 * we check that an individual a exists conforming to L for all outgoing edges.
		 * @param skeleton
		 * @param L
		 * @param constraints
		 * @param tau
		 * @return
		 * @throws JRDFStoreException
		 * @throws ExecutionException 
		 */
		public boolean satifiable(Map<Term, Set<Atom>> constraints) throws JRDFStoreException, ExecutionException {
			for (Term s : skeleton.vertexSet()) {
				if (s instanceof RootVariable) {
					if(!skeleton.outgoingEdgesOf(s).isEmpty() && !existRootVariable( L, constraints, s))
						return false;
				} else {
					if ((s instanceof Individual) && constraints.get(s) != null && !constraints.get(s).isEmpty()) {
						String query = SoundChecker.SW_PREFIX + reasoner.getPrefix() + EoloPrefixes.getDeclaration() + " SELECT  <a> WHERE { ";
						for (Atom constraint : constraints.get(s))  {
							 query +=  constraint.getArgument(0)+ " " + constraint.getDLPredicate()  + " "+ s+ " . \n";
						}
						query += "}";
						if (!reasoner.getCache().constraint().get(query))
							return false;
					}
					for (Term t : skeleton.vertexSet()) {
						if (skeleton.containsEdge(s, t)) {
							DefaultEdge e = skeleton.getEdge(s, t);
							if (!isLReachable(tau.get(s), tau.get(t), L.get(e)))
								return false;
						}	
					}
				}
			}
			return true;
		}
		

		private boolean existRootVariable(Map<DefaultEdge, Set<String>> L, Map<Term, Set<Atom>> constraints, 	Term rootVariable) throws JRDFStoreException {
			String query  = SoundChecker.SW_PREFIX + reasoner.getPrefix() + EoloPrefixes.getDeclaration() + 
					"SELECT " +rootVariable + " WHERE {" +
					rootVariable + " rdf:type " + EoloPrefixes.IND_PREFIX + "Ind .\n";
			if (constraints.containsKey(rootVariable)) {
				for (Atom constraint : constraints.get(rootVariable)) {
					query +=  constraint.getArgument(0)+ " " + constraint.getDLPredicate()  + " "+ rootVariable+ " . \n";
				}
			}
			query += "}";

			TupleIterator tupleIterator = null;
			try {
				tupleIterator = reasoner.getStore().compileQuery(query);
				for (tupleIterator.open(); tupleIterator.isValid(); tupleIterator.getNext()) {
					String datalog_s = tupleIterator.getGroundTerm(0).toString();
					for (Term t : skeleton.vertexSet()) {
						boolean allSatisfied = true;
						if (skeleton.containsEdge(rootVariable, t)) {
							DefaultEdge e = skeleton.getEdge(rootVariable, t);
							if (!isLReachable(datalog_s, tau.get(t), L.get(e)))
								allSatisfied = false;
						}	
						if (allSatisfied)
							return true;
					}
				}
				return false;
			} finally {
				if (tupleIterator != null)
					tupleIterator.dispose();
			}
		}
		
		private boolean isLReachable(String start, String end, Set<String> L) throws JRDFStoreException {
			if (L==null || L.isEmpty())
				return (new DatalogReachability(reasoner.getStore(), start)).marked(end);
			if (!reasoner.getHierarchy().allTransitive(L)) {
				String ruleBody = "";
				for (String r : L) {
					String role = "<" + EoloPrefixes.DIRECT_URI + URIUtils.shorten(r) +">";
					ruleBody = ruleBody + start + " " + role + " " + end + " .";
				}
				String query = "select <a> WHERE { " + ruleBody + "}";
				
				if(reasoner.entailsBCQ(query))
					return true;
				return false;
			} 			
			DatalogReachability reachability = new DatalogReachability(reasoner.getStore(), start, L);
			return reachability.marked(end);
		}	
	}
	
	public String toString() {
		return this.skeleton.toString();
	}

}
