package uk.ac.ox.eolo.reasoner;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.semanticweb.HermiT.model.Atom;
import org.semanticweb.HermiT.model.Role;
import org.semanticweb.HermiT.model.Term;
import org.semanticweb.owlapi.model.OWLOntology;

import uk.ac.ox.cs.JRDFox.JRDFStoreException;
import uk.ac.ox.cs.JRDFox.Prefixes;
import uk.ac.ox.cs.JRDFox.store.DataStore;
import uk.ac.ox.cs.JRDFox.store.DataStore.UpdateType;
import uk.ac.ox.cs.JRDFox.store.Parameters;
import uk.ac.ox.cs.JRDFox.store.TupleIterator;
import uk.ac.ox.eolo.URIUtils;
import uk.ac.ox.eolo.clausifier.OntologyClausifier2;
import uk.ac.ox.eolo.model.ConjunctiveQuery;
import uk.ac.ox.eolo.model.EoloPrefixes;
import uk.ac.ox.eolo.model.ExtendedConjunctiveQuery;
import uk.ac.ox.eolo.model.Result;
import uk.ac.ox.eolo.model.RoleHierarchy;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;


/**
 * 
 * @author 
 *
 */
public class EOLO {

	
	private String dec_prefix;
	private DataStore rdfFox;
	private RoleHierarchy roleISA;
	private Cache cache;
	

	protected EOLO() {}

	public static EOLO createReasoner(OWLOntology shoelOntology, File dataFile, File ruleFile, boolean transform) throws Exception {
		String onto_iri =shoelOntology.getOntologyID().getDefaultDocumentIRI()==null ? "urn:hermit:kb" : shoelOntology.getOntologyID().getDefaultDocumentIRI().toString();
		if (transform) {
			OntologyClausifier2.transformOntology(shoelOntology, dataFile, ruleFile);
			RoleHierarchy isa = RoleHierarchy.populate(shoelOntology);
			System.out.println("finished clausifying");
			shoelOntology = null;
			return createReasoner(dataFile, ruleFile, isa, onto_iri);
		} else {
			RoleHierarchy isa = RoleHierarchy.populate(shoelOntology);
			return createReasoner(dataFile, ruleFile, isa, onto_iri);
		}
	}

	public void dispose() {
		rdfFox.dispose();
	}

	public DataStore getStore() {
		return rdfFox;
	}

	public static EOLO createReasoner(File dataFile, File ruleFile,
			RoleHierarchy isa, String IRI) throws Exception {
		EOLO reasoner = new EOLO();
		reasoner.roleISA = isa;
		reasoner.cache = reasoner.new Cache();
		reasoner.cache.initOntologyCache();
		reasoner.dec_prefix = "PREFIX " + EoloPrefixes.ONTO_PREFIX + " <" + IRI + "#>\n"; 
		System.out.println("initialising the data store");
		reasoner.rdfFox = new DataStore(DataStore.StoreType.NarrowParallelHead);
		reasoner.rdfFox.setNumberOfThreads(2);
		System.out.println("adding the data file");
		reasoner.rdfFox.importTurtleFile(dataFile);
		System.out.println("number of facts: "
				+ reasoner.rdfFox.getTriplesCount());
		System.out.println("importing the rules.. ");
		reasoner.rdfFox.importRules(new File[] { ruleFile });
		System.out.println("materialising.. ");
		reasoner.rdfFox.applyReasoning();
		System.out.println("clearing rules and making facts explicit");
		reasoner.rdfFox.clearRulesAndMakeFactsExplicit();
		System.out.println("Adding facts over aux predicate..");
		reasoner.addAuxAssertions();
		System.out.println(StoreStatistics.computeStatistics(reasoner.getStore()));
		return reasoner;
	}
	
	
	/**
	 * For each individual u occurring in the materialisation such that u is an auxiliary individual
	 * and no individual c\in\indnames exists such that c \approx u, we add an assertion Aux(u).
	 * Morever, we also materialise a predicate AuxReach such that AuxReach(u',u) is true
	 * if and only if individuals u_0,.., u_m exist with u_0 = u' such that for each
	 * i\in\interval{1}{m}, we have u_i is auxiliary and for some role R we have
	 * d_R(u_{i-1}, u_i) is in the materialisation.
	 * @throws Exception
	 */
	private void addAuxAssertions() throws Exception {
		String prefix = EoloPrefixes.getDeclaration() + SoundChecker.SW_PREFIX + getPrefix();
		String query = prefix + "SELECT ?x WHERE {?x rdf:type owl:Thing.}";
		PrintWriter writer = null;
		TupleIterator tupleIterator = null;
		File dataFile = File.createTempFile("auxdata", "ttl");
		dataFile.deleteOnExit();
		try {
			writer = new PrintWriter(new BufferedWriter(new FileWriter(
					dataFile, false)));
			Parameters params = new Parameters();
			params.m_expandEquality = false;
			tupleIterator = rdfFox.compileQuery(query, new Prefixes(), params);
			for (long multiplicity = tupleIterator.open(); multiplicity != 0; multiplicity = tupleIterator.getNext()) {
				String groundTerm = tupleIterator.getGroundTerm(0).toString();
				if (isRealAnonymous(groundTerm)) {
					Set<String> selfRoles = new HashSet<String>();
					for (String r : this.roleISA.getRoles()) {
						if (hasSelfEdge(groundTerm, r))
							 selfRoles.add(r);
					}
					this.cache.addAuxIndividual(groundTerm, selfRoles);
					writer.println(groundTerm + " <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <"
							+ EoloPrefixes.AUX_URI + "Aux" + "> .");
				}
			}
		} finally {
			if (writer != null)
				writer.close();
			if (tupleIterator != null)
				tupleIterator.dispose();
		}
		this.rdfFox.importTurtleFiles(new File[] { dataFile },
				UpdateType.ScheduleForAddition);
		this.rdfFox.applyReasoning(true);
		System.out.println("Number of facts after addition: "
				+ this.rdfFox.getTriplesCount());
		System.out.println("clearing rules and making facts explicit");
		this.rdfFox.clearRulesAndMakeFactsExplicit();
	}
	
	private boolean hasSelfEdge(String groundTerm, String r) throws JRDFStoreException {
		String prefix = EoloPrefixes.getDeclaration() + SoundChecker.SW_PREFIX + getPrefix();
		String query = prefix + "SELECT <a> WHERE { " + groundTerm + " rdf:type " + " " + 
		EoloPrefixes.SELF_PREFIX + URIUtils.shorten(r) + " .}";
		if (entailsBCQ(query)) {
			return true;
		}
		return false;
	}

	public String getPrefix()  {
		return dec_prefix;
	}
	
	public ResultCollector answerCQ(ConjunctiveQuery q_pref) throws Exception {
		return answerCQ(q_pref, new Configuration());
	}
	
	private String toString(List<String> answer) {
		String result = "";
		for (String c : answer) {
			result += c + ", ";
		}
		return result.substring(0, result.lastIndexOf(','));
	}
		
	public ResultCollector answerCQ(ConjunctiveQuery q_pref, Configuration cf)
			throws Exception {
		ResultCollector answers = new ResultCollector();
		PrintWriter writer = null;
		if (cf.outputFile != null) {
			writer = new PrintWriter(new FileOutputStream(cf.outputFile), true);
		}
		Set<List<String>> distinctSet = null;
		if (cf.distinct){
			distinctSet = new HashSet<List<String>>();
		}
		ExtendedConjunctiveQuery qext = ExtendedConjunctiveQuery.computeExtension(q_pref); 
		Parameters params = new Parameters();
		params.m_expandEquality = true;
		TupleIterator tupleIterator = null;
		cache.initQueryCache();
		ThreadPoolExecutor exec = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors()/2, Runtime.getRuntime().availableProcessors()/2, 
				1000, TimeUnit.MILLISECONDS, new ArrayBlockingQueue<Runnable>(10));
		exec.setRejectedExecutionHandler(new ThreadPoolExecutor.CallerRunsPolicy());
        CompletionService<Result> service =    new ExecutorCompletionService<Result>(exec);		
		try {
			tupleIterator = rdfFox.compileQuery(qext.toString(), new Prefixes(), params); 
			long counter =0;
			for (tupleIterator.open(); tupleIterator.isValid(); tupleIterator.getNext()) {
				service.submit(new SoundChecker(this, qext, computeTauPrime(qext,	tupleIterator)));				
				counter++;
			}
			for(int tasksHandled=0;tasksHandled < counter;tasksHandled++){
                try {
                    Result res = service.take().get();
                    if (res.isSound() && (!cf.distinct || !distinctSet.contains(res.getAnswer()))) {
                    	answers.addSound();
                    	if (writer != null) 
                    		writer.println(toString(res.getAnswer()));
                    } else 
                    	answers.addUnSound();
                    answers.addTime(res.getTimeRequired());
                    answers.increaseChoices(res.getNondeterministic());
                } catch (InterruptedException e) {
                    System.out.println("Error Interrupted exception");
                    e.printStackTrace();
                } catch (ExecutionException e) {
                    e.printStackTrace();
                    System.out.println("Error get() threw exception");
                    return null;
                }
            }
			return answers;
		} finally {
			distinctSet = null;
			if (tupleIterator != null)
				tupleIterator.dispose();
			if (exec != null)
				exec.shutdownNow();
			if (writer != null)
				writer.close();
		}
	}

	public boolean entailsBCQ(String query) throws JRDFStoreException {
		Parameters params = new Parameters();
		params.m_expandEquality = false;
		TupleIterator tupleIterator = null;
		try {
			tupleIterator = rdfFox.compileQuery(query, new Prefixes(), params); 
			if (tupleIterator.open() > 0) 
				return true;
			return false;
		} catch (JRDFStoreException e) {
			e.printStackTrace();
			throw e;
		} finally {
			tupleIterator.dispose();
		}
	}
	
	
	/**
	 * Given ExtendedConjunctiveQuery qext and the substitution tau_prime,
	 * it returns the tuple of individuals to which tau_prime maps
	 * the answer variables of q.
	 * @param qext
	 * @param tau_prime
	 * @return
	 * @throws JRDFStoreException
	 */
	protected List<String> constructAnswer(ExtendedConjunctiveQuery qext,
			Map<Term, String> tau_prime) throws JRDFStoreException {
		List<String> answer = new ArrayList<String>();
		for (Term t : qext.getRealAnswerTerms())
			answer.add(tau_prime.get(t));
		return answer;
	}
	

	
	
	public Cache getCache() {
		return this.cache;
	}
	
	/**
	 * Given an ExtendedConjunctiveQuery qext and a TupleIterator,
	 * returns the substitution obtained from the current row in the tuple iterator.
	 * @param qext
	 * @param tupleIterator
	 * @return
	 * @throws JRDFStoreException
	 */
	private Map<Term, String> computeTauPrime(ExtendedConjunctiveQuery qext,
			TupleIterator tupleIterator) throws JRDFStoreException {
		Map<Term, String> tau_prime = new HashMap<Term, String>(qext.getNumberOfTerms());
		for (int i = 0; i < qext.getNumberOfAnswerTerms(); i++)
			tau_prime.put(qext.getAnswerTerm(i), tupleIterator.getGroundTerm(i).toString());
		for (Term t : qext.getIndividualsInBody())
			tau_prime.put(t, t.toString());
		return tau_prime;
	}
	
	/**
	 * Given an ExtendedConjunctiveQuery qext and a TupleIterator,
	 * returns the substitution obtained from the current row in the tuple iterator.
	 * @param qext
	 * @param tupleIterator
	 * @return
	 * @throws JRDFStoreException
	 */
	protected List<String> computeAnswer(ExtendedConjunctiveQuery qext,
			Map<Term, String> tau) throws JRDFStoreException {
		List<String> list = new ArrayList<String>();
		for (Term x : qext.getRealAnswerTerms())
			 list.add(tau.get(x));
		return list;
	}

	

	/**
	 * Given a string datalog_t representing an individual occurring in the materialisation,
	 * this method checks whether the individual is auxiliary.
	 * @param datalog_t
	 * @return
	 */
	public static boolean isSyntacticAnonymous(String datalog_t) {
		if (datalog_t.contains(EoloPrefixes.AUX_URI))
			return true;
		return false;
	}
	
	/**
	 * Checks whether the current string represents an individual u in the materialisation
	 * such that u \in Aux_{\K}.
	 * @param t
	 * @return
	 * @throws JRDFStoreException
	 * @throws ExecutionException 
	 */
	public  boolean isRealAnonymous(String t) throws JRDFStoreException, ExecutionException {
		String query = EoloPrefixes.getDeclaration()
				+ SoundChecker.SW_PREFIX + dec_prefix
				+ "select  <a> where { "
				+ t
				+ " rdf:type <http://www.cs.ox.ac.uk/EOLO/eolo-ind.owl#Ind> . \n"
				+ " } ";
		return isSyntacticAnonymous(t) && !entailsBCQ(query);
	}
	
	/**
	 * Checks whether atom a is aux-simple w.r.t. tau.
	 * @param a
	 * @param tau
	 * @return
	 * @throws Exception
	 */
	public boolean isAuxSimple(Atom a, Map<Term, String> tau, Prefixes pref) throws Exception {
		if (a.getArity() < 2)
			return true;
		Role r = (Role) a.getDLPredicate();
		Term s = a.getArgument(0);
		Term t = a.getArgument(1);
		String datalog_s = tau.get(s);
		String datalog_t = tau.get(t);
		return (!s.equals(t))
				&& this.cache.isRealAnonymous(datalog_t)
				&& getHierarchy().isSimple(URIUtils.getRole(r, pref))
				&& ((!datalog_s.equals(datalog_t)) || (!isAuxSelfLoop(a,tau, pref)));
	}
	

	
	/**
	 * Checks whether atom a is good w.r.t. tau.
	 * @param a
	 * @param tau
	 * @return
	 * @throws Exception
	 */
	public boolean isGood(Atom a, Map<Term, String> tau, Prefixes pref) throws Exception {
		if (a.getArity() < 2)
			return true;
		return !this.cache.isRealAnonymous(tau.get(a.getArgument(1))) || 
				(a.getArgument(0).equals(a.getArgument(1)) && isAuxSelfLoop(a,tau, pref));
	}
	

	
	/**
	 * Checks whether the given binary atom R(s,t) is a self-loop w.r.t. \tau;
	 * that is, it checks whether \tau(s)=\tau(t) and SELF_R(\tau(t)) is in the materialisation.
	 * @param a
	 * @param tau
	 * @return
	 * @throws JRDFStoreException
	 * @throws ExecutionException 
	 */
	public boolean isAuxSelfLoop(Atom a, Map<Term, String> tau, Prefixes pref)
			throws JRDFStoreException, ExecutionException {
		if (a.getArity() <2)
			return true;
		String firstPosition = tau.get(a.getArgument(0));
		String secondPosition = tau.get(a.getArgument(1));
		String r = URIUtils.getRole((Role) a.getDLPredicate(), pref);
		return (firstPosition.equals(secondPosition)) && this.cache.isRealAnonymous(secondPosition) && this.cache.hasSelfEdge(secondPosition, r);
	}

	public class Cache {
			
			private LoadingCache<String, Boolean> cacheConstraint;
			
			private LoadingCache<String, Boolean> cacheRole;
			
			
			private HashMap<String, Set<String>> auxInds;
			
			
			public void initOntologyCache() {
				auxInds = new HashMap<String, Set<String>>();
			}
			
			public boolean hasSelfEdge(String secondPosition, String r) {
				if (auxInds.containsKey(secondPosition) && auxInds.get(secondPosition).contains(r))
					return true;
				return false;
			}

			public boolean isRealAnonymous(String datalog_t) {		
				return auxInds.containsKey(datalog_t);
			}

			public void initQueryCache() {
				cacheConstraint =  CacheBuilder.newBuilder()
					       .maximumSize(1000)
					       .recordStats()
					       .build(
					           new CacheLoader<String, Boolean>() {
					             public Boolean load(String query) throws JRDFStoreException {
					               return entailsBCQ(query);
					             }
					           });

				cacheRole =  CacheBuilder.newBuilder()
					       .maximumSize(1000)
					       .recordStats()
					       .build(
					           new CacheLoader<String, Boolean>() {
					             public Boolean load(String query) throws JRDFStoreException {
					               return entailsBCQ(query);
					             }
					           });
			}
			
			public void addAuxIndividual(String auxind, Set<String> roles) {
				auxInds.put(auxind, roles);
				
			}

			public void clear() {
				cacheConstraint.cleanUp();
				cacheRole.cleanUp();
			}
			
			public LoadingCache<String, Boolean>  constraint() {
				return this.cacheConstraint;
			}
			
			public LoadingCache<String, Boolean>  roles() {
				return this.cacheRole;
			}

		}

	public RoleHierarchy getHierarchy() {
		return this.roleISA;
	}
	
	

}

 