package uk.ac.ox.eolo.reasoner;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;

import uk.ac.ox.eolo.clausifier.exception.IllegalInputOntologyException;
import uk.ac.ox.eolo.cqparser.ConjunctiveQueryParser;
import uk.ac.ox.eolo.exception.IllegalInputQueryException;
import uk.ac.ox.eolo.model.ConjunctiveQuery;

public class Main {
	

	/**
	 * Display example application header.
	 * 
	 * @out OutputStream to which header should be written.
	 */
	public static void displayHeader(final OutputStream out) {
		String header = "EOLO REASONER\n";
		try {
			out.write(header.getBytes());
		} catch (IOException ioEx) {
			System.out.println(header);
		}
	}

	/**
	 * Write the provided number of blank lines to the provided OutputStream.
	 * 
	 * @param numberBlankLines
	 *            Number of blank lines to write.
	 * @param out
	 *            OutputStream to which to write the blank lines.
	 */
	public static void displayBlankLines(final int numberBlankLines,
			final OutputStream out) {
		try {
			for (int i = 0; i < numberBlankLines; ++i) {
				out.write("\n".getBytes());
			}
		} catch (IOException ioEx) {
			for (int i = 0; i < numberBlankLines; ++i) {
				System.out.println();
			}
		}
	}

	/**
	 * Print usage information to provided OutputStream.
	 * 
	 * @param applicationName
	 *            Name of application to list in usage.
	 * @param options
	 *            Command-line options to be part of usage.
	 * @param out
	 *            OutputStream to which to write the usage information.
	 */
	public static void printUsage(String applicationName, Options options,
			OutputStream out) {
		PrintWriter writer = new PrintWriter(out);
		HelpFormatter usageFormatter = new HelpFormatter();
		usageFormatter.printUsage(writer, 80, applicationName, options);
		writer.flush();
	}

	/**
	 * Write "help" to the provided OutputStream.
	 */
	public static void printHelp(Options options, int printedRowWidth,
			String header, String footer, int spacesBeforeOption,
			int spacesBeforeOptionDescription, boolean displayUsage,
			OutputStream out) {
		String commandLineSyntax = "java -jar EOLO.jar";
		PrintWriter writer = new PrintWriter(out);
		HelpFormatter helpFormatter = new HelpFormatter();
		helpFormatter.printHelp(writer, printedRowWidth, commandLineSyntax,
				header, options, spacesBeforeOption,
				spacesBeforeOptionDescription, footer, displayUsage);
		writer.flush();
	}


	/**
	 * Construct and provide GNU-compatible Options.
	 * 
	 * @return Options expected from command-line of GNU form.
	 */
	public static Options constructOptions() {
		final Options options = new Options();
		OptionBuilder.isRequired(true);
		OptionBuilder.hasArg(true);
		OptionBuilder.withDescription("path to the file contanining the SHOEL ontology");
		options.addOption(OptionBuilder.create("ontology"));
		OptionBuilder.isRequired(true);
		OptionBuilder.hasArg(true);
		OptionBuilder.withDescription("path to the file containing the CQ");
		options.addOption(OptionBuilder.create("query"));
		OptionBuilder.isRequired(false);
		OptionBuilder.hasArg(true);
		OptionBuilder.withDescription("path to file which will contain the certain answers to the query over the ontology");
		options.addOption(OptionBuilder.create("output"));
		return options;
	}
	
	public static void main(String[] args) {
		displayBlankLines(1, System.out);
		displayHeader(System.out);
		displayBlankLines(2, System.out);
		CommandLineParser cmdLineGnuParser = new GnuParser();
		Options gnuOptions = constructOptions();
		CommandLine commandLine;
		try {
			commandLine = cmdLineGnuParser.parse(gnuOptions, args);

			File queryFile = new File(commandLine.getOptionValue("query"));
			File ontologyFile = new File(commandLine.getOptionValue("ontology"));
			if (!ontologyFile.exists()) {
				System.err.println("Could not locate ontology");
				return;
			}
			if (!queryFile.exists()) {
				System.err.println("Could not locate query");
				return;
			}
			File ruleFile = File.createTempFile("rules", "dlog");
			ruleFile.deleteOnExit();
			File dataFile = File.createTempFile("data", "ttl");
			dataFile.deleteOnExit();
			System.out.println("loading ontology file.. ");
			OWLOntologyManager man = OWLManager.createOWLOntologyManager();
			OWLOntology onto = man.loadOntologyFromOntologyDocument(ontologyFile);
			System.out.println("performing datalog transformation, and materialisation");
			double start = System.currentTimeMillis();
			EOLO eolo = EOLO.createReasoner(onto, dataFile, ruleFile, true);
			double end = System.currentTimeMillis();
			System.out.println("...done. It took me " + ((end - start) / 1000)	+ " seconds.\n");
			System.out.println("parsing input query...");
			ConjunctiveQuery q = new ConjunctiveQueryParser(queryFile).parse();
			System.out.println("...done.\n");
			System.out.println("evaluating query..");
			start = System.currentTimeMillis();
			ResultCollector res = null;
			if (commandLine.hasOption("output")) {
				Configuration cf = new Configuration();
				cf.outputFile =new File(commandLine.getOptionValue("output"));
				res = eolo.answerCQ(q, cf);
			} else
				res = eolo.answerCQ(q);
			
			end = System.currentTimeMillis();
			System.out.println("..done in "
					+ Double.toString(((end - start) / 1000)) + " seconds. \n");
			System.out.println(res);
		} catch (ParseException parseException) {
			displayBlankLines(4, System.err);
			printHelp(constructOptions(), 80, "HELP", "", 5, 3, true,
					System.err);
			displayBlankLines(4, System.err);
			System.err.println("Encountered exception while parsing:\n"
					+ parseException.getMessage());
		} catch (OWLOntologyCreationException e) {
			e.printStackTrace();
		} catch (IllegalInputOntologyException e) {
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IllegalInputQueryException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} 
	}

}
