package uk.ac.ox.eolo.reasoner;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;

import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.HermiT.model.Atom;
import org.semanticweb.HermiT.model.AtomicRole;
import org.semanticweb.HermiT.model.Individual;
import org.semanticweb.HermiT.model.Role;
import org.semanticweb.HermiT.model.Term;
import org.semanticweb.HermiT.model.Variable;

import uk.ac.ox.cs.JRDFox.JRDFStoreException;
import uk.ac.ox.cs.JRDFox.Prefixes;
import uk.ac.ox.cs.JRDFox.store.Parameters;
import uk.ac.ox.cs.JRDFox.store.TupleIterator;
import uk.ac.ox.eolo.URIUtils;
import uk.ac.ox.eolo.model.ConjunctiveQuery;
import uk.ac.ox.eolo.model.EoloPrefixes;
import uk.ac.ox.eolo.model.ExtendedConjunctiveQuery;
import uk.ac.ox.eolo.model.Result;
import uk.ac.ox.eolo.model.RootVariable;
import uk.ac.ox.eolo.model.Skeleton;
import uk.ac.ox.eolo.model.Skeleton.Labelling;
import uk.ac.ox.eolo.reasoner.exception.ConstraintException;

public class SoundChecker implements Callable<Result> {
	
	
	public static final String SW_PREFIX = "PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>\n"
			+ "PREFIX owl: <http://www.w3.org/2002/07/owl#>\n";
	
	private EOLO reasoner;
	private ExtendedConjunctiveQuery q;
	private Map<Term, String> tau;
	
	private long nondeterministic =0;
	
	
	public SoundChecker(EOLO r, ExtendedConjunctiveQuery qext, Map<Term, String> m) {
		this.reasoner = r;
		this.q = qext;
		this.tau = m;
	}
	
	@Override
	public Result call() throws Exception {
		return isSound();
	}
	
	/**
	 * Checks whether tau is sound w.r.t. q and the materialisation.
	 * @param q
	 * @param tau
	 * @return
	 * @throws Exception
	 */
	public  Result isSound() throws Exception {
		try {
			long start = System.currentTimeMillis();
			if(!answerVariablesMappedToNamed()) {
				return getResult(false, start);
			}
			if (!q.containsBinaryAtoms()) {
				return getResult(true,start);
			}
			EqualityConstraint sim = new EqualityConstraint();
			sim.computeConstraint();
			if (!sim.satisfied()) {
				return getResult(false,start);
			}
			if (isAuxSimpleCyclic(createSimpleGraph(this.q))) {
				return getResult(false,start);
			}
			boolean done = true;
			for (Atom a : q.getAtoms())
				if (!reasoner.isGood(a, tau, q.getPrefixes()) && !reasoner.isAuxSimple(a,tau, q.getPrefixes())) 
					done =false;
			if (done)  {
				return getResult(true,start);
			}
			ConjunctiveQuery q_sim = sim.simplifyQuery();
			return getResult((new VariableRenaming(q_sim)).exists(),start);
		} finally {
			this.reasoner = null;
			this.q = null;
			this.tau = null;
		}
	}

	private Result getResult(boolean isSound, long start) {
		Result result = new Result();
		result.setSound(isSound);
		result.setNondeterministic(nondeterministic);
		result.setTimeRequired(System.currentTimeMillis()-start);
		List<String> tuple = new ArrayList<String>(this.q.getNumberOfRealAnswerTerms());
		for (Term x : this.q.getRealAnswerTerms()){
			if (this.tau.containsKey(x))
				tuple.add(tau.get(x));
		}
		result.setAnswer(tuple);
		return result;
	}
	
	
	/**
	 * Checks whether all the answer variables in the original query
	 * are mapped to named individuals from the ontology.
	 * @param q_ext
	 * @param tau_prime
	 * @return
	 */
	public boolean answerVariablesMappedToNamed() {
		for (Term t : q.getRealAnswerTerms())
			if (EOLO.isSyntacticAnonymous(tau.get(t)))
				return false;
		return true;
	}
	

	
	/**
	 * Checks whether the graph obtained by adding an edge for each aux-simple atom R(s,t)\in q
	 * is cyclic.
	 * @param q
	 * @param tau
	 * @return
	 * @throws Exception
	 */
	public boolean isAuxSimpleCyclic(DirectedGraph<Term, DefaultEdge> auxGraph)
			throws Exception {
		return (new CycleDetector<Term, DefaultEdge>(auxGraph)).detectCycles();
	}


	private DirectedGraph<Term, DefaultEdge> createSimpleGraph(
			ConjunctiveQuery q) throws Exception {
		DirectedGraph<Term, DefaultEdge> auxGraph = new DefaultDirectedGraph<Term, DefaultEdge>(DefaultEdge.class);
		for (Term queryTerm : q.getTerms()) {
			auxGraph.addVertex(queryTerm);
		}
		for (Atom a : q.getAtoms()) {
			if (a.getArity() == 2 && !(a.getArgument(1) instanceof Individual) 	&& reasoner.isAuxSimple(a,tau, q.getPrefixes()))
				auxGraph.addEdge(a.getArgument(0), a.getArgument(1));
		}
		return auxGraph;
	}
	
	class EqualityConstraint {

		protected Map<Term, Set<Term>> sim;

		
		public EqualityConstraint() {
			sim = new HashMap<Term, Set<Term>>();
		}
		

		/**
		 * Checks whether all constraints are satisfied by tau.
		 * @return
		 * @throws ConstraintException
		 * @throws JRDFStoreException 
		 */
		public boolean satisfied() throws ConstraintException, JRDFStoreException {
			String body ="";
			for (Term s : q.getTerms())
				for (Term t : q.getTerms())
					if (areConstrained(s, t) && !tau.get(s).equals(tau.get(t))) {
						body += tau.get(s) + " owl:sameAs " + tau.get(t) + " . ";
					}
			if (body.isEmpty())
				return true;
			String prefix = EoloPrefixes.getDeclaration() + SW_PREFIX + reasoner.getPrefix();
			String query = prefix + "SELECT <a> WHERE { " + body + " }";
			return  reasoner.entailsBCQ(query);
		}

		
		/**
		 * Returns the query q_\sim.
		 * @param q
		 * @return
		 * @throws Exception 
		 * @throws ExecutionException 
		 * @throws JRDFStoreException 
		 */
		public ConjunctiveQuery simplifyQuery() throws JRDFStoreException, ExecutionException, Exception {
			Map<Term, Term> sub = new HashMap<Term, Term>();
			List<Term> indTerms = new ArrayList<Term>();
			for (Term t : q.getTerms()) {
				Term sim_repr = getSimRepresentative(t);
				if (t instanceof Variable && reasoner.getCache().isRealAnonymous(tau.get(sim_repr)))
					sub.put(t, getSimRepresentative(t));
				else {
					indTerms.add(t);
				}
			}
			computeRepresentatives(indTerms, sub);
			return q.applySubstitution(sub);
		}
		
		private void computeRepresentatives(List<Term> indTerms, Map<Term, Term> sub) throws Exception {
			if (indTerms != null && indTerms.size() >0) {
				String prefix = EoloPrefixes.getDeclaration() + SW_PREFIX + reasoner.getPrefix();
				String head = " ";
				String body = " ";
				int counter =0;
				for (Term i : indTerms) {
					head += " ?ind_" +  counter;
					body += "?ind_" +  counter + " owl:sameAs " + tau.get(i) + " . ";
					counter++;
				}
				String query = prefix + " SELECT " + head + " WHERE { " + body + " } ";
				Parameters params = new Parameters();
				params.m_expandEquality = false;
				TupleIterator tupleIterator = null;
				try {
					tupleIterator = reasoner.getStore().compileQuery(query, new Prefixes(), params);
					if (tupleIterator.open() == 0)
						throw new Exception("datalog term  does not have a representative");
					for (int i =0; i < counter; i++) {
						Individual ind = Individual.create(URIUtils.removeBrackets(tupleIterator.getGroundTerm(i).toString()));
						sub.put(indTerms.get(i), ind);
						tau.put(ind, ind.toString());
					}
				} finally {
					if (tupleIterator != null)
						tupleIterator.dispose();
				}
			}
		}
		

		/**
		 * Computes the relation \sim.
		 * @throws Exception
		 */
		public void computeConstraint() throws Exception {
			for (Term t : q.getTerms()) {
				addSingletonClass(t);
			}
			deriveForkConstraints();
		}

		
		protected void addSingletonClass(Term t) {
			Set<Term> eqclass = new HashSet<Term>();
			eqclass.add(t);
			sim.put(t, eqclass);
		}

		protected boolean areConstrained(Term s, Term t)
				throws ConstraintException {
			Term sRepresentative = getSimRepresentative(s);
			Term tRepresentative = getSimRepresentative(t);
			if (sRepresentative == null || tRepresentative == null) {
				throw new ConstraintException("Cannot identify terms " + s
						+ " and " + t);
			}
			return sRepresentative.equals(tRepresentative);
		}

		protected void constrain(Term s, Term t)
				throws ConstraintException {
			Term sRepresentative = getSimRepresentative(s);
			Term tRepresentative = getSimRepresentative(t);
			if (sRepresentative == null || tRepresentative == null) {
				throw new ConstraintException("Cannot identify terms " + s
						+ " and " + t);
			}
			if (!sRepresentative.equals(tRepresentative)) {
				sim.get(sRepresentative).addAll(sim.get(tRepresentative));
				sim.remove(tRepresentative);
			}
		}

		protected Term getSimRepresentative(Term s) {
			if (sim.containsKey(s))
				return s;
			for (Term key : sim.keySet()) {
				if (sim.get(key).contains(s))
					return key;
			}
			return null;
		}



		/**
		 * Derives all the constraints using the fork rule.
		 * @throws Exception
		 */
		protected void deriveForkConstraints() throws Exception {
			boolean newDerivedConstraints = true;
			while (newDerivedConstraints) {
				newDerivedConstraints = false;
				for (Atom a1 : q.getAtoms())
					for (Atom a2 : q.getAtoms()) {
						if (a1.getArity() >1 && a2.getArity() >1 && reasoner.isAuxSimple(a1,tau, q.getPrefixes()) && reasoner.isAuxSimple(a2,tau, q.getPrefixes())) {							
							String datalog_term = tau.get(a1.getArgument(1));
							if (areConstrained(a1.getArgument(1),a2.getArgument(1))
									&& !areConstrained(	a1.getArgument(0), a2.getArgument(0))) {
								if (reasoner.getCache().isRealAnonymous(datalog_term)) {
									constrain(a1.getArgument(0), a2.getArgument(0));
									newDerivedConstraints = true;
								}
							}
						}
					}
			}
		}
	}
	
	
	class VariableRenaming {
		private ConjunctiveQuery q_sim;
		public VariableRenaming(ConjunctiveQuery q) {
			this.q_sim = q;
		}
		/**
		 * Returns true if there exists a variable renaming sigma satisfying
		 * all conditions in steps 4--17 in Algorithm 1.
		 * @param q_sim
		 * @param tau
		 * @param q
		 * @return
		 * @throws Exception
		 * @throws JRDFStoreException
		 */
		public boolean exists() throws Exception,
				JRDFStoreException {
			Enumerator<Term> renEnumerator = new Enumerator<Term>(initEnumerator());
			for (Map<Term, Term> sigma : renEnumerator) {
				nondeterministic++;
				ConjunctiveQuery sigma_q = q_sim.applySubstitution(sigma);
				DirectedGraph<Term, DefaultEdge> simpleGraph = createSimpleGraph(sigma_q);
				if (isValidRenaming(sigma) && !isAuxSimpleCyclic(simpleGraph)) {//checking the conditions in Definition 9.
					boolean simpleOrGood = true;
					for (Atom a : sigma_q.getAtoms()) 
						if (!reasoner.isGood(a,tau, sigma_q.getPrefixes()) && !reasoner.isAuxSimple(a,tau, sigma_q.getPrefixes())) {
							simpleOrGood= false;
						}
					if(simpleOrGood || (new SkeletonEnumerator(sigma_q, simpleGraph).exists()))
						return true;
				}
			}
			return false;
		}
		
		/**
		 * Checks that for all terms s,t such that \sigma(s) = t, we have \sigma(t) =t.
		 * @param m
		 * @return
		 */
		private boolean isValidRenaming(Map<Term, Term> m) {
			for (Term x : m.keySet()) {
				Term y = m.get(x);
				if (!y.equals(m.get(y)))
					return false;
			}
			return true;
		}
		
		private List<Pair<Term, List<Term>>> initEnumerator() {
			List<Pair<Term, List<Term>>> list = new ArrayList<Pair<Term, List<Term>>>();
			for (Term t : q_sim.getTerms()) {
				if (t instanceof Variable) {
					List<Term> tList = new ArrayList<Term>();
					for (Term u : q_sim.getTerms()) {
						if ((u instanceof Variable)
								&& tau.get(t).equals(tau.get(u)))
							tList.add(u);
					}
					list.add(new Pair<Term, List<Term>>(t, tList));
				}
			}
			return list;
		}
	}

	
	
	class SkeletonEnumerator {
		private ConjunctiveQuery sigma_q;
		private DirectedGraph<Term, DefaultEdge> simpleGraph;

		SkeletonEnumerator(ConjunctiveQuery q, DirectedGraph<Term, DefaultEdge> simple) {
			this.sigma_q =q;
			this.simpleGraph = simple;
		}

		/**
		 * Returns true if there exists a skeleton satisfying all conditions in steps 5--17.
		 * @param sigma_q
		 * @param tau
		 * @return
		 * @throws Exception
		 * @throws JRDFStoreException
		 */
		boolean exists() throws Exception,
		JRDFStoreException {
			addRootVariables();
			Enumerator<Term> tEdges = new Enumerator<Term>(initEnumerator());
			for (Map<Term, Term> tEdge : tEdges) {
				nondeterministic++;
				Skeleton skeleton = Skeleton.create(simpleGraph, reasoner, sigma_q, tau);
				for (Term key : tEdge.keySet()) 
					skeleton.addEdge(tEdge.get(key), key);
				if (skeleton.isForest()) {
					if (new LabellingEnumerator(sigma_q, skeleton).exists())
						return true;
				}
			}
			return false;
		}	

		private void addRootVariables() throws Exception {
			for (Term t : sigma_q.getTerms()) 
				if (t instanceof Variable)
					simpleGraph.addVertex(RootVariable.create(t.toString() + "_ROOT"));
		}

		private List<Pair<Term, List<Term>>> initEnumerator() {
			List<Pair<Term, List<Term>>> list = new ArrayList<Pair<Term, List<Term>>>();
			for (Term t : simpleGraph.vertexSet()) {
				if (!(t instanceof RootVariable) && (t instanceof Variable)	&& simpleGraph.incomingEdgesOf(t).isEmpty())
					list.add(new Pair<Term, List<Term>>(t, new ArrayList<Term>(simpleGraph.vertexSet())));
			}
			return list;
		}
	}
	
	
	class LabellingEnumerator {
		
		private ConjunctiveQuery sigma_q;
		private Skeleton skeleton;
		
		LabellingEnumerator(ConjunctiveQuery q, Skeleton s) {
			this.sigma_q =q;
			this.skeleton =s;
		}
		
		/**
		 * Checks whether there exist a sequence of nondeterministic guesses so that steps 6--17 are satisfied.
		 * @param sigma_q
		 * @param skeleton
		 * @param tau
		 * @return
		 * @throws Exception
		 */
		boolean exists() throws Exception {
			Enumerator<Atom> atomEnumerator = new Enumerator<Atom>(initEnumerator()); 
			Map<Term, Set<Atom>> constraints = new HashMap<Term, Set<Atom>>();
			for (Map<Atom, Atom> atomMap : atomEnumerator) {
				Labelling L = skeleton.createLabelling();
				nondeterministic++;
				boolean satisfiesAllConditions = true;
				for (Atom a : atomMap.keySet()) {
					Atom b = atomMap.get(a);
					if (satisfiesAllConditions) {
						Term s = b.getArgument(0);
						Term t = b.getArgument(1);
						String p = URIUtils.getRole((Role) b.getDLPredicate(), sigma_q.getPrefixes());
						String query = EoloPrefixes.getDeclaration() + SW_PREFIX + reasoner.getPrefix() + "SELECT  <a> WHERE {" + 
						tau.get(s) + " <" + p + "> "+ tau.get(t) + " .}";
						if (reasoner.getCache().roles().get(query)) { // check entailment from Line 8 of the Algorithm
							if (skeleton.containsEdge(s, t) || reasoner.getHierarchy().isTransitive(p)) { // check in line 9
								if (skeleton.reaches(s, t)) {// conditional statement in Line 10
									L.put(p, skeleton.getPath(s,t));
								}
								else {
									Term root = skeleton.getRoot(t);
									if (root == null) {
										throw new Exception("no root found in skeleton");
									}
									Individual ind = Individual.create(URIUtils.removeBrackets(tau.get(s)));
									Set<Atom> atoms = (constraints.containsKey(root)) ? constraints.get(root)	: new HashSet<Atom>();
									atoms.add(Atom.create(b.getDLPredicate(), new Term[]{ind,root}));
									constraints.put(root, atoms);
									L.put(p, skeleton.getPath(root,t));
								}
							} else
								satisfiesAllConditions = false;
						} else
							satisfiesAllConditions = false;
					}
				}
				if(satisfiesAllConditions && L.satifiable(constraints))  {
					return true;
				}
			}
			return false;
		}
		

		private List<Pair<Atom, List<Atom>>> initEnumerator() throws Exception {
			List<Pair<Atom, List<Atom>>> list = new ArrayList<Pair<Atom, List<Atom>>>();
			for (Atom a : sigma_q.getAtoms()) {
				if (a.getArity()>1 && !(reasoner.isGood(a, tau, sigma_q.getPrefixes()) || reasoner.isAuxSimple(a,tau, sigma_q.getPrefixes()))) {
					Role r = (Role) a.getDLPredicate();
					ArrayList<Atom> subAtoms = new ArrayList<Atom>();
					for (String p : reasoner.getHierarchy().subRoles(URIUtils.getRole(r, sigma_q.getPrefixes()))) {
						Term[] terms = new Term[] { a.getArgument(0),a.getArgument(1) };
						subAtoms.add(Atom.create(AtomicRole.create(p), terms));
					}
					list.add(new Pair<Atom, List<Atom>>(a, subAtoms));
				}
			}
			return list;
		}
	}


	


	
}
